/* boeser Hack (c) 1996 by Martin Denn <mdenn@mathematik.uni-kl.de */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/../../drivers/scsi/sg.h>

#define MAXELEM 7
#define MAXTRY 10
#define MAXREPLYLEN 255

struct al_status
{
  unsigned int slot[MAXELEM];
  unsigned int allow[MAXELEM];
};


#define SCSI_OFF sizeof(struct sg_header)
static unsigned char buffer[10*1024];      /* SCSI command buffer */
int fd;                               /* SCSI device/file descriptor */

/* process a complete scsi cmd. Use the generic scsi interface. */
static int handle_scsi_cmd(unsigned cmd_len,         /* command length */
                           unsigned in_size,         /* input data size */
                           unsigned char *i_buff,    /* input buffer */
                           unsigned out_size,        /* output data size */
                           unsigned char *o_buff     /* output buffer */
                           )
{
  int status = 0;
  struct sg_header *sg_hd;

   /* safety checks */
   if (!cmd_len) return -1;            /* need a cmd_len != 0 */
   if (!i_buff) return -1;             /* need an input buffer != NULL */
#ifdef SG_BIG_BUFF
   if (SCSI_OFF + cmd_len + in_size > SG_BIG_BUFF) return -1;
   if (SCSI_OFF + out_size > SG_BIG_BUFF) return -1;
#else
   if (SCSI_OFF + cmd_len + in_size > 4096) return -1;
   if (SCSI_OFF + out_size > 4096) return -1;
#endif

   if (!o_buff) out_size = 0;

   /* generic scsi device header construction */
   sg_hd = (struct sg_header *) buffer;
   sg_hd->reply_len   = SCSI_OFF + out_size;
   sg_hd->twelve_byte = cmd_len == 12;
#if     0
   sg_hd->pack_len    = SCSI_OFF + cmd_len + in_size; /* not necessary */
   sg_hd->pack_id;     /* not used */
   sg_hd->other_flags; /* not used */
#endif

   memcpy(buffer + SCSI_OFF, i_buff, cmd_len + in_size);
   /* send command */
   status = write( fd, buffer, SCSI_OFF + cmd_len + in_size );
   if ( status < 0 || status != SCSI_OFF + cmd_len + in_size ||
                      sg_hd->result ) {
       /* some error happened */
       fprintf( stderr, "write(generic) result = 0x%x cmd = 0x%x\n", 
                   sg_hd->result, i_buff[SCSI_OFF] );
       perror("");
       return status;
   }
     
   if (!o_buff) o_buff = i_buff;       /* buffer pointer check */

   /* retrieve result */
   status = read( fd, buffer, SCSI_OFF + out_size);
   if ( status < 0 || status != SCSI_OFF + out_size || sg_hd->result ) {
       /* some error happened */
       fprintf( stderr, "read(generic) result = 0x%x cmd = 0x%x\n",
                    sg_hd->result, o_buff[SCSI_OFF] );
       fprintf( stderr, "read(generic) sense "
               "%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x\n", 
                               sg_hd->sense_buffer[0],         sg_hd->sense_buffer[1],
               sg_hd->sense_buffer[2],         sg_hd->sense_buffer[3],
               sg_hd->sense_buffer[4],         sg_hd->sense_buffer[5],
               sg_hd->sense_buffer[6],         sg_hd->sense_buffer[7],
               sg_hd->sense_buffer[8],         sg_hd->sense_buffer[9],
               sg_hd->sense_buffer[10],        sg_hd->sense_buffer[11],
               sg_hd->sense_buffer[12],        sg_hd->sense_buffer[13],
               sg_hd->sense_buffer[14],        sg_hd->sense_buffer[15]);
       perror("");
   }
   /* Look if we got what we expected to get */
   if (status == SCSI_OFF + out_size) status = 0; /* got them all */
   memcpy(o_buff, buffer+SCSI_OFF, out_size);

   return status;  /* 0 means no error */
}

#define INQUIRY_CMD     0x12
#define INQUIRY_CMDLEN  6

/* request vendor brand and model */
void Inquiry ( unsigned char *s )
{
  unsigned char cmdblk [ INQUIRY_CMDLEN ] = 
          { INQUIRY_CMD,  /* command */
                      0,  /* lun/reserved */
                      0,  /* page code */
                      0,  /* reserved */
   (MAXREPLYLEN & 0xff),  /* allocation length */
                      0 };/* reserved/flag/link */

  if (handle_scsi_cmd(sizeof(cmdblk), 0, cmdblk, 
                      (MAXREPLYLEN & 0xff), s )) {
      fprintf( stderr, "Inquiry failed\n" );
      exit(2);
  }
}

#define READ_ELEMENT_STATUS_CMD     0xb8
#define READ_ELEMENT_STATUS_CMDLEN  12

void ReadElementStatus ( unsigned char *s )
{
  unsigned char cmdblk [ READ_ELEMENT_STATUS_CMDLEN ] = 
      { READ_ELEMENT_STATUS_CMD,  /* command */
                              0,  /* lun/reserved */
                              0,  /* Starting Element Address, MSB */
                              1,  /* Starting Element Address, LSB */
                              0,  /* Number of Elements, MSB */
                              7,  /* Number of Elements, LSB */
                              0,  /* reserved */
     ((MAXREPLYLEN>>16) & 0xff),  /* allocation length, MSB */
      ((MAXREPLYLEN>>8) & 0xff),  /* allocation length */
           (MAXREPLYLEN & 0xff),  /* allocation length, LSB */
                              0,  /* reserved */
                              0 };/* reserved/flag/link */

  if (handle_scsi_cmd(sizeof(cmdblk), 0, cmdblk, 
                      MAXREPLYLEN, s )) {
      fprintf( stderr, "Read Element Status failed\n" );
      exit(2);
  }
}

char *elementtype[]=
{
   "",
   "Medium Transport Element",
   "Storage Element",
   "Import Export Element",
   "Data Transfer Element"
};

void print_RES ( unsigned char *s )
{
   int elem, no_of_el, byte_cnt, d_byte_cnt, desc_len, i, j, type;
   
    
   printf("first element     : %d\n", (s[0]<<8)+(s[1]));
   no_of_el=(s[2]<<8)+(s[3]);
   printf("no. of elements   : %d\n", no_of_el);
   byte_cnt=(s[5]<<16)+(s[6]<<8)+(s[7]);
   i=0;
   while ( i<byte_cnt )
   {
     printf("============================================\n");
     type=s[8+i]; i++;
     printf("element type      : %s\n", elementtype[type]);
     i++; /* reserved */
     desc_len = (s[8+i]<<8)+(s[8+i+1]); i+=2;
     i++; /* reserved */
     d_byte_cnt=(s[8+i]<<16)+(s[8+i+1]<<8)+(s[8+i+2]); i+=3;
     j=0;
     while ( j<d_byte_cnt )
     {
       printf("--------------------------------------------\n");
       elem = (s[8+i+j]<<8)+(s[8+i+j+1]); j+=2;
       printf("element address   : %d\n", elem);
       if ((s[8+i+j] & 0x08)==0x08) printf("access allowed\n"); 
              else printf("access denied\n");
       if ((s[8+i+j] & 0x04)==0x04) 
         printf("additional sense code: %02xh, qualifier: %02xh\n",
           s[8+i+j+2], s[8+i+j+3]);
       if ((s[8+i+j] & 0x01)==0x01) printf("element contains media\n");
       else printf("element contains no media\n");
       j++;
       j+=3; /* reserved, additional sense code */
       if (type==4)
       {
         if ((s[8+i+j] & 0x80)==0x80) 
           printf("changer and primary device on separate buses\n");
         if ((s[8+i+j] & 0x10)==0x10) 
           printf("primary device has SCSI bus address %d\n",s[8+i+j+1]);
         if ((s[8+i+j] & 0x20)==0x20) 
           printf("primary device has SCSI LUN %d\n",(s[8+i+j] & 0x07));
       }
       j+=3; /* reserved or SCSI addr/LUN */
       if ((s[8+i+j] & 0x80)==0x80)
       {
         printf("source storage element address: %d ",
             (s[8+i+j+1]<<8)+(s[8+i+j+2]));
         if ((s[8+i+j] & 0x40)==0x40) printf(" (inverted)\n");
         else printf("\n");
       }
       j+=3; /* above */
       j+=4; /* reserved */
       if ((j % desc_len) != 0)
       {
           printf("Vendor Unique : ");
           while ((j % desc_len) != 0) { putchar(s[8+i+j]); j++; }
           printf("\n");
       }
     }
     i+=d_byte_cnt;
  }
}
    
int get_al_status ( unsigned char *s, struct al_status *as )
{
   int elem, byte_cnt, d_byte_cnt, desc_len, i, j, type;
    
    
    byte_cnt=(s[5]<<16)+(s[6]<<8)+(s[7]);
    i=0;
    while ( i<byte_cnt )
    {
      type=s[8+i]; i++;
      i++; /* reserved */
      desc_len = (s[8+i]<<8)+(s[8+i+1]); i+=2;
      i++; /* reserved */
      d_byte_cnt=(s[8+i]<<16)+(s[8+i+1]<<8)+(s[8+i+2]); i+=3;
      j=0;
      while ( j<d_byte_cnt )
      {
        elem = (s[8+i+j]<<8)+(s[8+i+j+1]); j+=2;
	if (elem>MAXELEM)
	{
          return 1;
	}
        if ((s[8+i+j] & 0x08)==0x08) as->allow[elem-1]=1;
	else as->allow[elem-1]=1;
        if ((s[8+i+j] & 0x01)==0x01) as->slot[elem-1]=1;
        else as->slot[elem-1]=0;
        j++;
        j+=3; /* reserved, additional sense code */
        j+=3; /* reserved or SCSI addr/LUN */
        if ((s[8+i+j] & 0x80)==0x80)
        {
          as->slot[elem-1]=(s[8+i+j+1]<<8)+(s[8+i+j+2]);
          if (as->slot[elem-1]>0) (as->slot[elem-1])--;
        }
        j+=3; /* above */
        j+=4; /* reserved */
        while ((j % desc_len) != 0) j++;
      }
      i+=d_byte_cnt;
    }
    return 0;
}
     
#define MOVE_MEDIUM_CMD 0xa5
#define MOVE_MEDIUM_CMDLEN 12

void MoveMedium ( int i, int j )
{
  static unsigned char cmdblk [MOVE_MEDIUM_CMDLEN] = {
        MOVE_MEDIUM_CMD, /* command */
                      0, /* lun/reserved */
                      0, /* Transport Element Address, MSB */
                      0, /* Transport Element Address, LSB */
                      0, /* Source Address, MSB */
                      0, /* Source Address, LSB */
                      0, /* Destination Address, MSB */
                      0, /* Destination Address, LSB */
                      0, /* reserved */
                      0, /* reserved */
                      0, /* reserved */
                      0};/* reserved */

  cmdblk[4]=(unsigned char) ((i & 0x0000ff00)>>8);
  cmdblk[5]=(unsigned char) (i & 0x000000ff);
  cmdblk[6]=(unsigned char) ((j & 0x0000ff00)>>8);
  cmdblk[7]=(unsigned char) (j & 0x000000ff);
  

  if (handle_scsi_cmd(sizeof(cmdblk), 0, cmdblk, 
                            0, NULL)) {
      fprintf (stderr, "Move Medium failed\n");
      exit(2);
  }
}

#define LOAD_CMD 0x1b
#define LOAD_CMDLEN 6

void EjectMag ()
{
  static unsigned char cmdblk [LOAD_CMDLEN] = {
               LOAD_CMD, /* command */
                      0, /* lun/reserved */
                      0, /* Media ID, MSB */
                      0, /* Media ID, LSB */
                      0, /* LSB: load=1, unload=0 */
                   0x80};/* MSB: unload mag=1 */
  
  if (handle_scsi_cmd(sizeof(cmdblk), 0, cmdblk, 
                            0, NULL)) {
      fprintf (stderr, "Eject Magazine failed\n");
      exit(2);
  }
}

void print_status( struct al_status *al )
{
  printf("status report:\n");
  printf("                    1 %c | %c 4\n", 
    (al->slot[1]?'x':'-'),  (al->slot[4]?'x':'-'));
  printf("                    2 %c | %c 5\n", 
    (al->slot[2]?'x':'-'),  (al->slot[5]?'x':'-'));
  printf("                    3 %c | %c 6\n", 
    (al->slot[3]?'x':'-'),  (al->slot[6]?'x':'-'));
  if (al->slot[0]) printf("loaded cartridge is from slot %d.\n", al->slot[0]);
  else printf("no cartridge loaded.\n");
}
     

void unload ( int i, struct al_status *al )
{
  if (al->slot[0] == 0)
  {
    fprintf(stderr, "no cartdrige loaded\n");
    return;
  } 
  if (i==0) i=al->slot[0];
  if ((i<1) || (i>6))
  {
    fprintf(stderr, "no such cartridge (%d).\n", i);
    exit(1);
  }
  if ((!al->allow[i]) || (!al->allow[0]))
  {
    fprintf(stderr, "unload cartridge not allowed.\n");
    exit(1);
  }
  else MoveMedium( 1, i+1);
}

void load( int i, struct al_status *al )
{
  if ((i<1) || (i>6))
  {
    fprintf(stderr, "no such cartridge (%d).\n", i);
    exit(1);
  }
  if (al->slot[0] == i) 
  {
    fprintf(stderr, "cartridge from slot %d already loaded\n",i);
    return;
  } 
  else if (al->slot[i] == 0)
  {
    fprintf(stderr, "no cartridge in slot %d\n",i);
    exit(1);
  } 
  else if ( al->slot[0] != 0 ) unload(0, al);
  if ((!al->allow[i]) || (!al->allow[0]))
  {
    fprintf(stderr, "load cartridge not allowed.\n");
    exit(1);
  }
  else MoveMedium( i+1, 1);
}

void eject ( struct al_status *al )
{
  if  (al->slot[0] != 0) unload(0, al);
  EjectMag();
}




void usage( char *s )
{
  fprintf( stderr, "-------------------------------------------------------------\n");
  fprintf( stderr, "%s sends commands to an HP C1553A autoloader device\n", s);
#ifdef GENDEVICE
  fprintf( stderr, "USAGE: %s [-f generic-SCSI-device] command\n", s);
#else /* GENDEVICE */
  fprintf( stderr, "USAGE: %s -f generic-SCSI-device command\n", s);
#endif
  fprintf( stderr, "commands:\n");
  fprintf( stderr, "  info        get full info about autoloader\n");
  fprintf( stderr, "  status      show slot status\n");
  fprintf( stderr, "  load n      load cartridge from slot n\n");
  fprintf( stderr, "  unload [n]  unload cartridge to original slot [to slot n]\n");
  fprintf( stderr, "  eject       eject magazine\n");
  fprintf( stderr, "-------------------------------------------------------------\n");
  exit(1);
} 

void main( int argc, char **argv )
{
  unsigned char action, *file, *progname;
  unsigned char s[MAXREPLYLEN];
  struct al_status al;
  int i, nr, try;
   
  progname=argv[0];
  if (argc==1) usage(progname);

#ifdef GENDEVICE
  file=GENDEVICE;
#else /*GENDEVICE*/
  file=NULL;
#endif

  i=1; nr=0; action='\0';
  while (i<argc)
  {
    if (argv[i][0]=='-')
    {
      if (argv[i][1]=='f')
      {
         if (argv[i][2]!='\0')
         {
           file = &(argv[i][2]); i++;
         }
         else if (argc > i+1) file = argv[i+1];
         {
           file = argv[i+1]; i+=2;
         }
      }
      else
      {
        fprintf( stderr, "%s: Option -%c not supported\n",
      		   progname, argv[i][1]);
        usage(progname);
      }
    }
    else
    {
      if (action!='\0')
      {
        fprintf( stderr, "%s: too many commands\n", progname);
        usage(progname);
      }
      switch (argv[i][0])
      {
      case 'i':
      case 'e':
      case 's':
        action=argv[i][0]; i++;
        break;
      case 'l':
        if (argc>i+1) nr=atoi(argv[i+1]);
        if ((nr<1) || (nr>=MAXELEM))
        {
            fprintf( stderr, "%s: please specify cartridge to load\n",
      	       progname);
            exit(1);
        }
    	  action=argv[i][0]; i+=2;
        break;
      case 'u':
        if (argc>i+1) nr=atoi(argv[i+1]);
        if ((nr<1) || (nr>=MAXELEM))
        {
           nr=0; action=argv[i][0]; i++;
        }
        else
	{
	  action=argv[i][0]; i+=2;
	}
        break;
      default:
        fprintf( stderr, "%s: unknown command (%s)\n",
      		   progname, argv[i]);
        usage(progname);
      }
    }
  }
  if (file==NULL)
  {
    fprintf( stderr, "%s: please specify the SCSI device\n",
	     progname);
    exit(1);
  }
  fd = open(file, O_RDWR);
  if (fd < 0) {
    fprintf( stderr, "%s: could not open SCSI device (%s)\n",
	     progname, file);
    fprintf( stderr, "  note: You must be root to use the generic SCSI interface.\n");
    exit(1);
  }

  Inquiry(s);
  if ((strncmp(s+8,"HP",2)!=0) || (strncmp(s+16,"C1553A",6)!=0))
  {
    fprintf( stderr, "%s: %s is no HP C1553A autoloader device\n",
	     progname, file);
    exit(1);
  }

  try=0;
  ReadElementStatus(s);
  while ((get_al_status(s, &al)!=0) && (try<MAXTRY))
  {
    try++;
    sleep(5);
    ReadElementStatus(s);
  }
  if (try>=MAXTRY)
  {
    fprintf( stderr, "%s: timeout in reading element status\n", progname);
    exit(1);
  }
  
  switch (action)
  {
  case 'i':
    /* print some fields of the ReadElementStatus result */
    print_RES(s);
    break;
  case 's':
    /* print full status */
    print_status(&al);
    break;
  case 'l':
    /* load specified cartridge */
    load(nr, &al);
    break;
  case 'u':
    /* unload specified cartridge */
    unload(nr, &al);
    break;
  case 'e':
    /* eject magazine */
    eject(&al);
    break;
  default:
     usage(argv[0]);
  }
  close(fd);
}
